from __future__ import annotations as _annotations

from typing import TYPE_CHECKING as _TYPE_CHECKING
import re as _re

if _TYPE_CHECKING:
    from pyprotocol import Stringable


def format(
    text: Stringable,
    text_styles: int | str | list[int | str] = None,
    text_color: int | str | tuple = None,
    bg_color: int | str | tuple = None,
    reset: bool = True
):
    sequence = create_sequence(text_styles, text_color, bg_color)
    return apply_sequence(text, sequence, reset=reset)


def create_sequence(
    text_styles: int | str | list[int | str] = None,
    text_color: int | str | tuple = None,
    bg_color: int | str | tuple = None
):
    styles = []
    if text_styles:
        if isinstance(text_styles, (str, int)):
            text_styles = [text_styles]
        if not isinstance(text_styles, list):
            raise TypeError("styles must be a string, an integer, or a list of strings or integers")
        for text_style in text_styles:
            styles.append(text_style_code(text_style))
    if text_color:
        styles.append(color_code(text_color, bg=False))
    if bg_color:
        styles.append(color_code(bg_color, bg=True))
    if not styles:
        return reset_sequence()
    return create_sequence_from_code(styles)


def create_sequence_from_code(code: list[Stringable] | Stringable):
    if isinstance(code, (list, tuple)):
        code = ";".join([str(c) for c in code])
    return f"\033[{code}m"


def apply_sequence(text: Stringable, sequence: Stringable, reset: bool = True):
    return f"{sequence}{text}{reset_sequence() if reset else ''}"


def remove_sequence(text: str):
    """Remove ANSI escape sequences from a string.

    Parameters
    ----------
    text

    Notes
    -----
    ANSI escape codes start with the \\033 (\\x1b) escape character,
    followed by '[', then zero or more numbers separated by ';', and ending with a letter.

    Regex Details:
    - [0-?]*: This part matches zero or more characters in the range between 0 and ?.
    This covers all the numbers and semicolons that might be present in the escape sequence.
    For example, in the code \\x1b[31;42m, 31 and 42 are matched by this part.
    - [ -/]*: This is a sequence of characters that might appear in some of the ANSI sequences.
    It's more of a catch-all for certain sequences and may not be strictly necessary for many common sequences.
    But it ensures we catch even those rare ANSI codes.
    - [@-~]: Finally, ANSI escape sequences end with a character from @ to ~:
    @ A B C D E F G H I J K L M N O P Q R S T U V W X Y Z
    [ \\ ] ^ _ ` a b c d e f g h i j k l m n o p q r s t u v w x y z {.
    This part matches that ending character. This character typically indicates what action
    should be taken (e.g., change color, move cursor, clear screen, etc.).
    """
    return _re.compile(r'\x1b\[[0-?]*[ -/]*[@-~]').sub('', text)


def reset_sequence():
    return create_sequence_from_code(0)


def text_style_code(text_style: Stringable | int) -> str:
    text_style_map = {
        'normal': '0',
        'bold': '1',
        'faint': '2',
        'italic': '3',
        'underline': '4',
        'blink': '5',
        'blink_fast': '6',
        'reverse': '7',
        'conceal': '8',
        'strike': '9',
    }
    if isinstance(text_style, int):
        if text_style not in range(10):
            raise ValueError(f"Invalid style code: {text_style}")
        return str(text_style)
    text_style = str(text_style)
    if text_style not in text_style:
        raise ValueError(f"Invalid style name: {text_style}")
    return text_style_map[text_style]


def color_code(color: int | Stringable | tuple[int, int, int], bg: bool = False):
    color_code_map = {
        'black': 30,
        'red': 31,
        'green': 32,
        'yellow': 33,
        'blue': 34,
        'magenta': 35,
        'cyan': 36,
        'white': 37,
        'b_black': 90,
        'b_red': 91,
        'b_green': 92,
        'b_yellow': 93,
        'b_blue': 94,
        'b_magenta': 95,
        'b_cyan': 96,
        'b_white': 97,
    }

    int_range = (
        list(range(40, 48)) + list(range(100, 108)) if bg else
        list(range(30, 38)) + list(range(90, 98))
    )
    int_offset = 10 if bg else 0
    rgb_code = 48 if bg else 38
    if isinstance(color, int):
        if color not in int_range:
            raise ValueError(f"Invalid color code: {color}")
        return str(color)
    if isinstance(color, (tuple, list)):
        if len(color) != 3:
            raise ValueError(f"Invalid color tuple: {color}")
        if not all(isinstance(c, int) for c in color):
            raise ValueError(f"Invalid color tuple: {color}")
        if not all(c in range(256) for c in color):
            raise ValueError(f"Invalid color tuple: {color}")
        return f"{rgb_code};2;{';'.join([str(c) for c in color])}"
    color = str(color)
    if color not in color_code_map:
        raise ValueError(f"Invalid color name: {color}")
    return f"{color_code_map[color] + int_offset}"



"""Work with ANSI SGR (Select Graphic Rendition) parameters.

References
----------
- https://en.wikipedia.org/wiki/ANSI_escape_code#SGR_(Select_Graphic_Rendition)_parameters
"""

from ansi_sgr.sgr import (
    format,
    create_sequence,
    create_sequence_from_code,
    apply_sequence,
    remove_sequence,
    reset_sequence,
    text_style_code,
    color_code,
)

from ansi_sgr import element


from __future__ import annotations as _annotations

from typing import TYPE_CHECKING as _TYPE_CHECKING
import textwrap as _textwrap
import wcwidth as _wcwidth

import ansi_sgr as _sgr

if _TYPE_CHECKING:
    from typing import Literal
    from pyprotocol import Stringable


def admonition(
    title: Stringable,
    text: Stringable,
    emoji: Stringable | None = None,
    title_styles: int | str | list[int | str] = "bold",
    title_color: int | str | tuple = (255, 255, 255),
    title_bg_color: int | str | tuple = (70, 0, 0),
    title_margin_top: int | None = None,
    title_margin_bottom: int | None = None,
    title_margin_left: int | None = None,
    title_margin_right: int | None = None,
    title_align: Literal["left", "right", "center"] = "left",
    text_styles: int | str | list[int | str] = None,
    text_color: int | str | tuple = None,
    text_bg_color: int | str | tuple = None,
    text_margin_top: int | None = None,
    text_margin_bottom: int | None = None,
    text_margin_left: int | None = None,
    text_margin_right: int | None = None,
    text_align: Literal["left", "right", "center"] = "left",
    char_top: Stringable | None = None,
    char_bottom: Stringable | None = "━",
    char_left: Stringable | None = "┃",
    char_right: Stringable | None = "┃",
    char_top_left: Stringable | None = "┏",
    char_top_right: Stringable | None = "┓",
    char_bottom_left: Stringable | None = "┗",
    char_bottom_right: Stringable | None = "┛",
    line_width: int = 70,
):
    title = str(title)
    if emoji:
        title = f"{emoji} {title}"
    title_section = block(
        text=title,
        text_styles=title_styles,
        text_color=title_color,
        bg_color=title_bg_color,
        char_top=char_top,
        char_bottom="",
        char_left=char_left,
        char_right=char_right,
        char_top_left=char_top_left,
        char_top_right=char_top_right,
        margin_top=title_margin_top,
        margin_bottom=title_margin_bottom,
        margin_left=title_margin_left or text_margin_left,
        margin_right=title_margin_right or text_margin_right,
        line_width=line_width,
        align=title_align,
    )
    body_section = block(
        text=text,
        text_styles=text_styles,
        text_color=text_color,
        bg_color=text_bg_color,
        char_top="",
        char_bottom=char_bottom,
        char_left=char_left,
        char_right=char_right,
        char_bottom_left=char_bottom_left,
        char_bottom_right=char_bottom_right,
        margin_top=text_margin_top,
        margin_bottom=text_margin_bottom,
        margin_left=text_margin_left or title_margin_left,
        margin_right=text_margin_right or title_margin_right,
        line_width=line_width,
        align=text_align,
    )
    return f"{title_section}\n{body_section}"


def block(
    text: Stringable,
    text_styles: int | str | list[int | str] = None,
    text_color: int | str | tuple = None,
    bg_color: int | str | tuple = None,
    char_top: Stringable | None = "━",
    char_bottom: Stringable | None = "━",
    char_left: Stringable | None = "┃",
    char_right: Stringable | None = "┃",
    char_top_left: Stringable | None = "┏",
    char_top_right: Stringable | None = "┓",
    char_bottom_left: Stringable | None = "┗",
    char_bottom_right: Stringable | None = "┛",
    margin_top: int | None = None,
    margin_bottom: int | None = None,
    margin_left: int | None = None,
    margin_right: int | None = None,
    line_width: int = 50,
    align: Literal["left", "right", "center"] = "left",
):
    lines = _textwrap.wrap(
        text=str(text),
        width=line_width,
        expand_tabs=True,
        tabsize=4,
        replace_whitespace=False,
        drop_whitespace=False,
    )
    if margin_left is None:
        margin_left = margin_right
    if margin_right is None:
        margin_right = margin_left
    if margin_top is None:
        margin_top = margin_bottom
    if margin_bottom is None:
        margin_bottom = margin_top
    if margin_top:
        lines = [" " * line_width] * margin_top + lines
    if margin_bottom:
        lines = lines + [" " * line_width] * margin_bottom
    formatted_lines = []
    for line in lines:
        screen_width = _wcwidth.wcswidth(line)
        char_width = len(line)
        extra_padding = char_width - screen_width
        width = line_width + extra_padding
        if align == "left":
            line = line.ljust(width)
        elif align == "right":
            line = line.rjust(width)
        else:
            line = line.center(width)
        formatted_lines.append(
            inline(
                line,
                text_styles=text_styles,
                text_color=text_color,
                bg_color=bg_color,
                char_left=char_left,
                char_right=char_right,
                margin_left=margin_left,
                margin_right=margin_right,
            )
        )
    if char_top is None:
        char_top = char_bottom or ""
    if char_bottom is None:
        char_bottom = char_top or ""
    total_line_width = line_width + (margin_left or 0) + (margin_right or 0)
    if char_top:
        line_top = f"{char_top_left}{char_top * total_line_width}{char_top_right}"
        formatted_lines.insert(0, line_top)
    if char_bottom:
        line_bottom = f"{char_bottom_left}{char_bottom * total_line_width}{char_bottom_right}"
        formatted_lines.append(line_bottom)
    return "\n".join(formatted_lines)


def inline(
    text: Stringable,
    text_styles: int | str | list[int | str] = None,
    text_color: int | str | tuple = None,
    bg_color: int | str | tuple = None,
    char_left: Stringable | None = None,
    char_right: Stringable | None = None,
    margin_left: int | None = None,
    margin_right: int | None = None,
):
    sequence = _sgr.create_sequence(text_styles, text_color, bg_color)
    text = str(text)
    if char_left is None:
        char_left = char_right or ""
    if char_right is None:
        char_right = char_left or ""
    if margin_left is None:
        margin_left = margin_right
    if margin_right is None:
        margin_right = margin_left
    if margin_left:
        text = " " * margin_left + text
    if margin_right:
        text = text + " " * margin_right
    text_box = _sgr.apply_sequence(text, sequence, reset=True)
    return f"{char_left}{text_box}{char_right}"
